"""
This module contains the NatsJob abstractions.

NatsJob manages the interactions with a specific job (defines the status, progress, etc.).
It is associated with a specific object store in the NATS server, for storing input/output data.

@author: Pierre Dellenbach
"""

import nats
import nats.aio
import nats.aio.subscription
import nats.aio.transport
import nats.errors
import nats.js
import nats.js.api
import nats.js.errors
import nats.js.object_store
import struct

from typing import Optional, List
import enum
import rich
import random
import string

from .nats_context import NatsContext


def _generate_id(length: int = 6, chars=string.ascii_lowercase + string.digits) -> str:
    """
    Generates a random id for a job
    """
    return "".join(random.SystemRandom().choice(chars) for _ in range(length))


class JobStatus(enum.Enum):
    """
    Defines the status of a job.

    Useful for querying the status of a job in the NATS context
    """

    CREATED = "CREATED"
    RUNNING = "RUNNING"
    CRASHED = "CRASHED"
    INTERRUPTED = "INTERRUPTED"
    COMPLETED = "COMPLETED"
    INVALID_STATE = "INVALID_STATE"

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def from_str(status: str) -> "JobStatus":
        key = status.upper()
        if key not in JobStatus.__members__:
            return JobStatus.INVALID_STATE
        return JobStatus[key]


class NatsJob:
    """
    A class for managing a job within the NATS context
    """

    def __init__(self, job_id: str, nats_context: NatsContext):
        self.job_id = job_id
        self.nats_context = nats_context
        self.object_store = None
        assert job_id is not None, "job_id must not be None"

    @staticmethod
    async def create(
        nats_context: NatsContext,
        job_id: Optional[str] = None,
        create_if_not_exists: bool = False,
    ) -> "NatsJob":
        """
        Creates a job within the NATS context

        If the job does not exist, it creates the job and sets the status to CREATED
        """
        assert nats_context._is_initialized, "NatsContext must be initialized"
        if job_id is None:
            assert (
                create_if_not_exists
            ), "job_id must be provided if create_if_not_exists is False"
            job_id = _generate_id()

        job = NatsJob(job_id, nats_context)
        await job._init_job(create_if_not_exists)
        return job

    @property
    def status_key(self):
        return "status"

    @property
    def progress_key(self):
        return "progress"

    @staticmethod
    def get_store_id(job_id: str) -> str:
        return f"job_{job_id}"

    @property
    def store_id(self):
        return self.get_store_id(self.job_id)

    @staticmethod
    async def list_jobs(nats_context: NatsContext) -> List[str]:
        job_ids = []
        for store in await nats_context.list_object_stores():
            if store.startswith("job_"):
                job_id = "_".join(store.split("_")[1:])
                job_ids.append(job_id)

        return job_ids

    async def _init_job(self, create_if_not_exists: bool = True):
        """
        Connects the job to the NATS context, and creates/loads an associated object store
        If the job does not exist, it creates the job and sets the status to CREATED
        """
        self.object_store = await self.nats_context.get_object_store(
            self.store_id, create_if_not_exists
        )
        status: JobStatus = None
        progress: float = None
        try:
            status = await self.object_store.get(self.status_key)
        except nats.js.errors.ObjectNotFoundError:
            status = None

        try:
            progress = await self.object_store.get(self.progress_key)
        except nats.js.errors.ObjectNotFoundError:
            progress = None

        if not status:
            await self.object_store.put(self.status_key, str(JobStatus.CREATED))
        if not progress:
            await self.object_store.put(self.progress_key, struct.pack("f", 0.0))

    async def status(self) -> JobStatus:
        """
        Returns the status of the job
        """
        assert self.object_store is not None, "Job not initialized"
        try:
            res: nats.js.object_store.ObjectStore.ObjectResult = (
                await self.object_store.get(self.status_key)
            )
            status_str = res.data.decode("utf-8")
        except nats.js.errors.ObjectNotFoundError:
            return JobStatus.INVALID_STATE

        return JobStatus.from_str(status_str)

    async def update_status(self, status: JobStatus):
        """
        Updates the status of the job
        """
        assert self.object_store is not None, "Job not initialized"
        status_value = str(status)
        await self.object_store.put(self.status_key, status_value)

    async def progress(self) -> float:
        """
        Returns the progress of the job
        """
        assert self.object_store is not None, "Job not initialized"
        try:
            res: nats.js.object_store.ObjectStore.ObjectResult = (
                await self.object_store.get(self.progress_key)
            )
            progress = struct.unpack("f", res.data)[0]
        except nats.js.errors.ObjectNotFoundError:
            return 0.0

        return progress

    async def update_progress(self, progress: float):
        """
        Updates the progress of the job
        """
        assert isinstance(progress, float), "Progress must be a float"
        assert progress >= 0.0 and progress <= 1.0, "Progress must be between 0 and 1"
        try:
            await self.object_store.put(self.progress_key, struct.pack("f", progress))
        except Exception as e:
            raise e

    async def delete_job(self) -> bool:
        """
        Deletes the job, and the associated object store(s)
        """
        try:
            res = await self.nats_context.delete_object_store(self.store_id)
            self.object_store = None
            return res
        except Exception as e:
            rich.print(f"[red]Error Deleting Job: {e}[/red]")
            raise e

    async def list_objects(self) -> List[str]:
        """
        Lists all objects in the object store of the job
        """
        obj_infos = await self.object_store.list(ignore_deletes=True)
        return [obj.name for obj in obj_infos]

    async def start_job(self):
        """
        Starts the job
        """
        status = await self.status()
        if status != JobStatus.CREATED:
            raise Exception("Job must be in CREATED state to start")
        await self.update_status(JobStatus.RUNNING)
        # Publish a message to the job's subject to start the job
        await self.nats_context.nats_client.publish(
            f"{self.jobs_subject_prefix(self.nats_context)}.{self.job_id}",
            self.job_id.encode("utf-8"),
        )
        await self.nats_context.nats_client.flush(timeout=1)
        rich.print(f"[green]Job {self.job_id} started[/green]")

    @staticmethod
    def jobs_subject_prefix(nats_context: NatsContext) -> str:
        return f"{nats_context.app_id}.job"



"""
This module defines the JobAPI class.

The JobAPI class is an abstract class to construct standardized Jobs APIs on top of NATS.

@author: Pierre Dellenbach
"""

import asyncio
from contextlib import asynccontextmanager
from fastapi import FastAPI, HTTPException, Depends, Query, status
from fastapi.security import HTTPBearer
import nats
import nats.aio
import nats.js.object_store
import os
from pydantic import BaseModel
import rich
from typing import Optional, Callable, List, Union


import chaps_nats


class _JobID(BaseModel):
    """
    Result of a job submission
    """

    job_id: str = "awshZ2"


class JobAPI:
    """
    An abstract class for an API which receives job submissions and either processes them (for rapid jobs)
    Or pushes them to a NATS queue for processing by workers.

    app_id is the unique identifier for the NATS workspace.
    It is important, as it is used to create the NATS object stores and queues

    Workers:
    app_id.workers.>worker_id (worker_id is the unique identifier for the worker)

    Queue:
    app_id.job_id.message
    """

    def __init__(
        self,
        app_id: str,
        nats_urls: List[str],
        api_prefix: str = "",
        app: Optional[FastAPI] = None,
        title: str = "TODO: Chaps Job API",
        description: str = "TODO: Add a description",
        *args,
        **kwargs,
    ):
        self.api_prefix = api_prefix
        self.app_id = app_id
        self.nats_urls = nats_urls

        # Create the FastAPI app
        self.app = app
        self.auth = None
        if self.app is None:
            # Create the FastAPI app using the default model expected by iaparc
            self.app = FastAPI(
                lifespan=self._get_lifespan(self),
                title=title,
                description=description,
                openapi_url="/docs/openapi.json",
                root_path=os.getenv(key="ROOT_URL", default=""),
            )
            self.auth = HTTPBearer()
        else:
            assert isinstance(self.app, FastAPI), "app must be an instance of FastAPI"

        self.add_routes()

    @staticmethod
    def _get_lifespan(_self: "JobAPI"):
        """
        Returns the lifespan of the FastAPI app
        """

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            # ============ Check Environment variables ============ #
            # ============ Load the flags database ============ #
            # Load from environment
            await _self.init_lifespan()
            yield
            await _self.close_lifespan()

        return lifespan

    def _make_nats_context(self) -> chaps_nats.NatsContext:
        context = chaps_nats.NatsContext(app_id=self.app_id, urls=self.nats_urls)
        return context

    async def init_lifespan(self):
        try:
            # Try to initialize the NATS context with a timeout
            nats_context = self._make_nats_context()
            await asyncio.wait_for(nats_context._init_context(), timeout=10)
        except Exception as e:
            rich.print("[b red]Error initializing NATS context. Caught exception[/]")
            rich.print(e)
            exit(1)

    async def close_lifespan(self):
        """
        Closes the context (typically resources like processes, connections, etc..)
        """
        pass

    def add_routes(self):
        """
        Adds API routes to the FastAPI app
        """
        # Submit job
        _submit_job_fun = self._job_submit_fun()
        if _submit_job_fun is not None:
            self.app.post(
                f"{self.api_prefix}/submit_job", tags=["Jobs"], response_model=_JobID
            )(_submit_job_fun)

        # Results
        _results_job_fun = self._job_results_fun()
        if _results_job_fun is not None:
            self.app.get(f"{self.api_prefix}/job_results/{{job_id}}", tags=["Jobs"])(
                _results_job_fun
            )

        # Status
        _jobs_status_fun = self._jobs_status_fun()
        if _jobs_status_fun is not None:
            self.app.get(f"{self.api_prefix}/job_status/{{job_id}}", tags=["Jobs"])(
                _jobs_status_fun
            )

        # Delete
        _job_delete_fun = self._job_delete_fun()
        if _job_delete_fun is not None:
            self.app.delete(f"{self.api_prefix}/delete_job/{{job_id}}", tags=["Jobs"])(
                _job_delete_fun
            )

        # List Jobs
        _job_lists_fun = self._job_lists_fun()
        if _job_lists_fun is not None:
            self.app.get(f"{self.api_prefix}/job_list", tags=["Jobs"])(_job_lists_fun)

    def _job_submit_fun(self) -> Optional[Callable]:
        return None

    def _job_results_fun(self) -> Optional[Callable]:
        return None

    def _job_delete_fun(self) -> Optional[Callable]:
        """
        Returns the default delete job function
        """

        async def delete_job(
            job_id: str,
            auth: HTTPBearer = Depends(self.auth),
        ):
            """
            Deletes a job
            """
            async with self._make_nats_context() as nats_context:
                try:
                    job = await chaps_nats.NatsJob.create(nats_context, job_id)
                    delete_job = await job.delete_job()
                    await asyncio.sleep(0)
                    return {"ok": delete_job}
                except Exception as e:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND,
                        detail={"message": "Job not found", "error": str(e)},
                    )

        return delete_job

    def _jobs_status_fun(self) -> Optional[Callable]:
        """
        Returns the default job status function
        """

        async def job_status(
            job_id: str,
            # TODO: Add a functionality to append / store messages for a job with_messages: bool = Query(
            #     False,
            #     description="Include the array of messages in the response (key **'messages'**).",
            # ),
            auth: HTTPBearer = Depends(self.auth),
        ):
            """
            ## Returns the status of a job
            """
            async with self._make_nats_context() as nats_context:
                try:
                    job = await chaps_nats.NatsJob.create(nats_context, job_id)
                except Exception as e:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND,
                        detail={"message": "Job not found", "error": str(e)},
                    )
                result = {
                    "status": await job.status(),
                    "progress": await job.progress(),
                }
                # TODO: Add a message mechanism
                return result

        return job_status

    def _job_lists_fun(self) -> Callable:
        """
        Returns the default job list function
        """

        async def info(
            auth: HTTPBearer = Depends(self.auth),
        ):
            """
            Get the list of jobs and their status
            """
            async with self._make_nats_context() as nats_context:
                jobs = await chaps_nats.NatsJob.list_jobs(nats_context)

                jobs_model = []
                for job_id in jobs:
                    job = await chaps_nats.NatsJob.create(nats_context, job_id)
                    try:
                        jobs_model.append(
                            {
                                "job_id": job.job_id,
                                "status": await job.status(),
                                "progress": await job.progress(),
                            }
                        )
                    except Exception as e:
                        continue

                return {"jobs": jobs_model}

        return info



"""
An abstract class for a worker which processes jobs from a NATS queue
"""

from abc import abstractmethod
import asyncio

import multiprocessing as mp
from multiprocessing import Value as V
import nats
import nats.aio
import nats.aio.client
import nats.errors
from typing import Dict, List, Optional
import rich

import chaps_nats


class NatsJobWorkerProcess(mp.Process):
    """
    A worker which processes jobs from a NATS queue

    >>> class MyWorker(NatsJobWorkerProcess):
    >>>     async def process_job(self, job : chaps_nats.NatsJob) -> bool:
    >>>         # Process the job
    >>>         print(f"Processing Job: {job.job_id}")
    >>>         return True
    >>>
    >>> if __name__ == "__main__":
    >>>     worker = MyWorker(nats_urls=["nats://localhost:4222"], app_id="MyWorker")
    >>>     worker.start()
    >>>     worker.join()

    """

    def __init__(
        self,
        nats_urls: List[str],
        queue_group: str = "",
        app_id: str = "",
        timeout: float = 0.1,
        *args,
        **kwargs,
    ):
        """
        Constructor
        :param nats_client: NatsClient
        :param subject: str
        :param queue_group: str
        """
        super().__init__(*args, **kwargs)
        self.nats_urls = nats_urls
        self.queue_group = queue_group
        self.app_id = app_id
        self._do_stop = V("b", False)
        self.timeout_s = timeout

    def stop(self):
        """
        Stops the worker
        """
        with self._do_stop.get_lock():
            self._do_stop.value = True

    def run(self):
        """
        Runs the worker
        """
        asyncio.run(self._run())

    async def init_context(self, nats_context: chaps_nats.NatsContext) -> Dict:
        """
        Initializes a context for the worker
        """
        return {}

    async def _run(self):
        """
        Runs the worker
        """
        async with chaps_nats.NatsContext(self.app_id, self.nats_urls) as nats_context:
            subject = chaps_nats.NatsJob.jobs_subject_prefix(nats_context) + ".>"
            sub = await nats_context.nats_client.subscribe(
                subject,
                queue=self.queue_group,
            )
            context = await self.init_context(nats_context)

            # Subscribe to the input subject
            while not self._do_stop.value:
                try:
                    await asyncio.sleep(0.001)

                    try:
                        msg = await sub.next_msg(timeout=self.timeout_s)
                    except asyncio.TimeoutError:
                        continue

                    job_id = msg.data.decode("utf-8")
                    rich.print(f"[green]Received Job: {job_id}[/green]")

                    try:
                        job = await chaps_nats.NatsJob.create(
                            nats_context, job_id, create_if_not_exists=False
                        )

                        # Check the status of the job
                        job_status = await job.status()
                        if job_status != chaps_nats.JobStatus.RUNNING:
                            rich.print(
                                f"[red]Job {job_id} is not in RUNNING state : {str(job_status)}[/red]"
                            )
                            continue

                        res = await self.process_job(job, context)
                        if not res:
                            await job.update_status(chaps_nats.JobStatus.CRASHED)
                            await job.update_progress(1.0)
                        else:
                            await job.update_status(chaps_nats.JobStatus.COMPLETED)
                            await job.update_progress(1.0)

                    except Exception as e:
                        rich.print(f"[red]Error Processing Job: {e}[/red]")
                        continue

                except asyncio.CancelledError:
                    with self._do_stop.get_lock():
                        self._do_stop.value = True
                    break

    @abstractmethod
    async def process_job(self, job: chaps_nats.NatsJob, context: Dict) -> bool:
        """
        Processes a job
        """
        raise NotImplementedError()
